/*
Fusion Comp Exporter for Premiere Pro v0.5
2018/x/x    - First version
2020/02/19  - Added support for sped up and slowed down clips (not time remapped).
2020/03/25  - Added support for time reversed clips.
2021/11/15  - Fixed frame rate bug. Handles offsets now the start frame.
2021/11/16  - Added support to export non trimmed loaders. Added support to export just timecode and frame number data.

Notes:
    - Does not export R3D RAW files' data.
    - Time reversed and non orginal speed clips won't export perfectly.

To do:
    > Fix reversed and time stretched clip handling...

Created by Arttu Rautio (aturtur)
Webpage http://aturtur.com
*/

$.runScript = {
    exportComp: function() {
        app.enableQE(); // Enable QE Dom
        // ----------------------------------------------------------------------------------------------------------------
        // Functions
        var getOSDivider = function() {
            if ($.os.indexOf("Win") != -1) {
                return "\\";
            } else {
                return "/";
            }
        }

        var splitByLast = function(text, separator) {
            // Split string by last found separator
            var index = text.lastIndexOf(separator);
            return [text.slice(0, index), text.slice(index + 1)]
        }

        var framesToTimeCode = function(frames, fps) {
            // Converts frames to time code
            var strF    = "00";
            var strS    = "00";
            var strM    = "00";
            var strH    = "00";

            var frameDuration = 1/fps;

            var f       = frames;
            var seconds = Math.floor(f * frameDuration);
            var minutes = Math.floor(seconds / 60);
            var hours   = Math.floor(minutes / 60);
            
            f = f % (1 / frameDuration);
            seconds = seconds % 60;
            minutes = minutes % 60;

            if(f < 10) {
                strF = "0" + f.toString();
            } else {
                strF = f.toString();
            }

            if(seconds < 10) {
                strS = "0" + seconds.toString();
            } else {
                strS = seconds.toString();
            }

            if(minutes < 10) {
                strM = "0" + minutes.toString();
            } else {
                strM = minutes.toString();
            }

            if(hours < 10) {
                strH = "0" + hours.toString();
            } else {
                strH = hours.toString();
            }

            var timeCode = strH+":"+strM+":"+strS+":"+strF;

            return timeCode;
        }

        var tickToFps = function(tick) {
            var fps = 0;
            switch(tick) {
                case 25401600000:                   // 10 fps
                    fps = 10;
                    break;
                case 21168000000:                   // 12 fps
                    fps = 12;
                    break;
                case 20321280000:                   // 12.5 fps
                    fps = 12.5;
                    break;
                case 16934400000:                   // 15 fps
                    fps = 15;
                    break;
                case 10594584000:                   // 23.973 fps
                    fps = 23.973;
                    break;
                case 10584000000:                   // 24 fps
                    fps = 24;
                    break;
                case 10160640000:                   // 25 fps
                    fps = 25;
                    break;
                case 8475667200:                    // 29.97 fps
                    fps = 29.97;
                    break;
                case 8467200000:                    // 30 fps
                    fps = 30;
                    break;
                case 5080320000:                    // 50 fps
                    fps = 50;
                    break;
                case 4237833600:                    // 59.94 fps
                    fps = 59.94;
                    break;
                case 4233600000:                    // 60 fps
                    fps = 60;
                    break;
                default:
                    fps = 25;                       // Default
            }
            return fps;
        }

        // ----------------------------------------------------------------------------------------------------------------
        // Common variables
        var seq = app.project.activeSequence;       // Get active sequence
        if (app.project.activeSequence == null) {
            alert("No sequence open!");
        }
        var seqName = seq.name;                     // Get sequence's name
        var resHeight = seq.frameSizeVertical;      // Get sequence's height resolution
        var resWidth = seq.frameSizeHorizontal;     // Get sequence's width resolution
        var trackCount = seq.videoTracks.numTracks; // Get video track count
        var tick = seq.timebase;                    // Get sequence's tick
        var fps = tickToFps(parseInt(tick));        // Convert tick to fps
        globalStart = globalStart - handles;        // Adjust global start frame

        // Initialize variables for later use
        var loaders      = [];                      // Initialize list for loader nodes
        var loaders_s    = [];                      //
        var timespeeds   = [];                      // Initialize list for time speed nodes
        var delays_s     = [];                      //
        var savers_s     = [];                      //
        var loaderCount  = 0;                       // Initialize loader count variable
        var fileFormat   = "";                      // Initialize string variable for file format type
        var error        = 0;                       // Initialize variable for error
        var atLeastOne   = 0;                       //
        var exporterData = ""                       // Initialize variable for data (text format)
        // ----------------------------------------------------------------------------------------------------------------
        // Run the main thing
        for (var k = 0; k< trackCount; k++) {       // Iterate through video tracks
            var track = seq.videoTracks[k];         // Get video track
            var clipCount = track.clips.numItems;   // Get clip count
            for (var i = 0; i < clipCount; i++) {   // Iterate through video clips
                var clip = track.clips[i];          // Get video clip
                if (clip.isSelected() == true) {    // If video clip is selected
                    atLeastOne = 1;                 // At least one selected video clip
                    var metadata = clip.projectItem.getProjectMetadata(); // Get clip's metadata
                    var timeCode = metadata.split("MediaStart>")[1].split("<")[0];          // Get media start in time code
                    var tcHH = parseInt(timeCode.split(":")[0]);                              // Hours (timeCode)
                    var tcMM = parseInt(timeCode.split(":")[1]);                              // Minutes (timeCode)
                    var tcSS = parseInt(timeCode.split(":")[2]);                              // Seconds (timeCode)
                    var tcFF = parseInt(timeCode.split(":")[3]);                              // Frames (timeCode)
                    var duration = metadata.split("MediaDuration>")[1].split("<")[0];       // Get media duration in time code                 
                    var HH = parseInt(duration.split(":")[0]);                              // Hours (duration)
                    var MM = parseInt(duration.split(":")[1]);                              // Minutes (duration)
                    var SS = parseInt(duration.split(":")[2]);                              // Seconds (duration)
                    var FF = parseInt(duration.split(":")[3]);                              // Frames (duration)
                    var fileName = clip.projectItem.getMediaPath();                         // Get file name
                    fileName = fileName.replace(/\\/g , "\\\\");                            // For Windows paths
                    fileName = "\""+fileName+"\"";                                          // 
                    var reversed = clip.isSpeedReversed()                                   // Check is video clip's speed reversed
                    var speed = clip.getSpeed()                                             // Get video clip's speed multiplier
                    var length = ((HH*3600)*fps)+((MM*60)*fps)+(SS*fps)+FF;                 // Calculate clip's length
                    var trimIn = Math.floor((clip.inPoint.ticks / tick) * speed);           // Calculate clip's in point
                    var trimOut= Math.floor(((clip.outPoint.ticks / tick) - 1) * speed);    // Calculate clip's out point
                    var fileExtension = splitByLast(fileName,".")[1];                       //
                    fileExtension = fileExtension.toLowerCase();                            //
                    fileExtension = fileExtension.replace(/[^\w\s]/gi, '')                  // Get file extension
                    var startFrame = splitByLast(fileName,"_")[1];                          //
                    startFrame = splitByLast(startFrame,".")[0];                            // Get start frame
                    if (reversed == true) {                                                 // If clip is time reversed
                        oldTrimIn = trimIn;                                                 //
                        trimIn = length - trimOut - 1;                                      // Calculate time reversed in point
                        trimOut = length - oldTrimIn - 1;                                   // Calculate time reversed out point
                    }
                    // *************************************************************************************************
                    // Check if there is room for handles
                    if (trimIn - handles < 0) {
                        alert("No enough space for handles!\nClip: " + fileName);
                        error = 1;
                    } else if (length - trimOut + handles > length) {
                        alert("No enough space for handles!\nClip: " + fileName);
                        error = 1;
                    }
                    // *************************************************************************************************
                    // File format stuff
                    switch(fileExtension){
                        case "r3d":
                            var formatId = "\"R3DFormat\"";
                            var sf = "";
                            var formatSpecific = "[\"Clip1.R3DFormat.ColorScience\"] = Input { Value = 1, },";
                            break;
                        case "mov":
                            var formatId = "\"QuickTimeMovies\"";
                            var sf = "";
                            var formatSpecific = "";
                            break;
                        case "mp4":
                            var formatId = "\"QuickTimeMovies\"";
                            var sf = "";
                            var formatSpecific = "";
                            break;
                        case "exr":
                            var formatId = "\"OpenEXRFormat\"";
                            var sf = "StartFrame = "+startFrame+",";
                            var formatSpecific = "[\"Clip1.OpenEXRFormat.RedName\"] = Input { Value = FuID { \"R\" }, },\
                    [\"Clip1.OpenEXRFormat.GreenName\"] = Input { Value = FuID { \"G\" }, },\
                    [\"Clip1.OpenEXRFormat.BlueName\"] = Input { Value = FuID { \"B\" }, },";
                            break;
                        case "dpx":
                            var formatId = "\"DPXFormat\"";
                            var sf = "StartFrame = "+startFrame+",";
                            var formatSpecific = "[\"Clip1.DPXFormat.SLogVersion\"] = Input { Value = FuID { \"SLog2\" }, },";
                            break;
                        case "png":
                            var formatId = "\"PNGFormat\"";
                            var sf = "StartFrame = "+startFrame+",";
                            var formatSpecific = "";
                            break;
                        case "jpg":
                            var formatId = "\"JpegFormat\"";
                            var sf = "StartFrame = "+startFrame+",";
                            var formatSpecific = "";
                            break;
                        case "tif":
                            var formatId = "\"TiffFormat\"";
                            var sf = "StartFrame = "+startFrame+",";
                            var formatSpecific = "";
                            break;
                        default:
                            break;
                    }
                    // *************************************************************************************************
                    // Loader --
                    var loaderid = loaderCount+parseInt(1);             // Get current loader ID
                    var loaderypos = loaderCount*parseInt(50);          // Get y-position for current loader
                    var loader = "\n\t\tLoader"+loaderid+" = Loader {\
                 Clips = {\
                    Clip {\
                        ID = \"Clip1\",\
                        Filename = "+fileName+",\
                        FormatID = "+formatId+",\
                        "+sf+"\
                        Length = "+length+",\
                        LengthSetManually = true,\
                        TrimIn = "+(trimIn-(handles*speed))+",\
                        TrimOut = "+(trimOut+(handles*speed))+",\
                        ExtendFirst = 0,\
                        ExtendLast = 0,\
                        Loop = 1,\
                        AspectMode = 0,\
                        Depth = 0,\
                        TimeCode = 0,\
                        GlobalStart = "+globalStart+",\
                        GlobalEnd = "+(((trimOut+(handles*speed))-(trimIn-(handles*speed)))+globalStart)+"\
                    }\
                },\
                Inputs = {\
                    [\"Gamut.SLogVersion\"] = Input { Value = FuID { \"SLog2\" }, },\
                    "+formatSpecific+"\
                },\
                ViewInfo = OperatorInfo { Pos = { 100, "+loaderypos+" } },\
            }";
                    // --------------------------------------------------------------------------------------------------
                    var loader_s = "\n\t\tLoader"+loaderid+" = Loader {\
                 Clips = {\
                    Clip {\
                        ID = \"Clip1\",\
                        Filename = "+fileName+",\
                        FormatID = "+formatId+",\
                        "+sf+"\
                        Length = "+length+",\
                        LengthSetManually = true,\
                        TrimIn = 0,\
                        TrimOut = "+(length-1)+",\
                        ExtendFirst = 0,\
                        ExtendLast = 0,\
                        Loop = 1,\
                        AspectMode = 0,\
                        Depth = 0,\
                        TimeCode = 0,\
                        GlobalStart = "+globalStart+",\
                        GlobalEnd = "+(globalStart+length-1)+"\
                    }\
                },\
                Inputs = {\
                    [\"Gamut.SLogVersion\"] = Input { Value = FuID { \"SLog2\" }, },\
                    "+formatSpecific+"\
                },\
                ViewInfo = OperatorInfo { Pos = { 100, "+loaderypos+" } },\
            }";
                    // *************************************************************************************************
                    // Create Time Speed node if necessary
                    if (exportMethod == 0) {
                        var secondNode = "Loader";
                    } else {
                        var secondNode = "TimeOffset";
                    }

                    if (reversed == true) { // If time reversed
                        speed = speed * -1.0
                    }
                    if (speed != 1) { // If speed is not 100%
                        var thirdNode = "TimeSpeed";
                        var timespeed = "\n\t\tTimeSpeed"+loaderid+" = TimeSpeed {\
            Inputs = {\
                Speed = Input { Value = "+speed+", },\
                Input = Input {\
                    SourceOp = \""+secondNode+loaderid+"\",\
                    Source = \"Output\",\
                },\
            },\
            ViewInfo = OperatorInfo { Pos = { 500, "+loaderypos+" } },\
        }";
                        timespeeds.push(timespeed); // Add time speed to time speeds list
                    } else {
                        var thirdNode = "TimeOffset";
                    }

                    // *************************************************************************************************
                    // Delay Time Speed nodes --
                    if (exportMethod == 2) { // If export method is set to Animate Savers 
                        var inFrame  = globalStart;
                        var outFrame = globalStart+((trimOut-trimIn)/speed)+(handles*2);


                        var delay_s = "\n\t\tTimeOffset"+loaderid+" = TimeSpeed {\
            Inputs = {\
                Speed = Input { Value = 1, },\
                Delay = Input { Value = -"+((trimIn-(handles*speed))/speed)+", },\
                Input = Input {\
                    SourceOp = \"Loader"+loaderid+"\",\
                    Source = \"Output\",\
                },\
            },\
            ViewInfo = OperatorInfo { Pos = { 250, "+loaderypos+" } },\
        }";
                        delays_s.push(delay_s); // Add delay node to the delay nodes list

                    // Savers nodes --
                        var saver_s = "\n\t\tSaver"+loaderid+" = Saver {\
            Inputs = {\
                Blend = Input {\
                    SourceOp = \"Saver"+loaderid+"Blend\",\
                    Source = \"Value\",\
                },\
                ProcessWhenBlendIs00 = Input { Value = 0, },\
                Clip = Input {\
                    Value = Clip {\
                        Filename = \"\",\
                        FormatID = \"OpenEXRFormat\",\
                        Length = 0,\
                        Saving = true,\
                        TrimIn = 0,\
                        ExtendFirst = 0,\
                        ExtendLast = 0,\
                        Loop = 1,\
                        AspectMode = 0,\
                        Depth = 0,\
                        GlobalStart = -2000000000,\
                        GlobalEnd = 0\
                    },\
                },\
                OutputFormat = Input { Value = FuID { \"OpenEXRFormat\" }, },\
                [\"Gamut.SLogVersion\"] = Input { Value = FuID { \"SLog2\" }, },\
                Input = Input {\
                    SourceOp = \""+thirdNode+loaderid+"\",\
                    Source = \"Output\",\
                },\
                [\"OpenEXRFormat.Depth\"] = Input { Value = 1, },\
                [\"OpenEXRFormat.Compression\"] = Input { Value = 6, },\
                [\"OpenEXRFormat.AlphaEnable\"] = Input { Value = 0, },\
                [\"OpenEXRFormat.ZEnable\"] = Input { Value = 0, },\
            },\
            ViewInfo = OperatorInfo { Pos = { 750, "+loaderypos+" } },\
        },\
        Saver"+loaderid+"Blend = BezierSpline {\
            SplineColor = { Red = 205, Green = 205, Blue = 205 },\
            KeyFrames = {\
                ["+(inFrame-1)+"] = { 0, LH = { "+(inFrame-1)+", 0 }, RH = { "+(inFrame-1)+", 0 }, Flags = { StepIn = true } },\
                ["+inFrame+"] = { 1, LH = { "+inFrame+", 0 }, RH = { "+inFrame+", 0 }, Flags = { StepIn = true } },\
                ["+outFrame+"] = { 1, LH = { "+outFrame+", 0 }, RH = { "+outFrame+", 1 }, Flags = { StepIn = true } },\
                ["+(outFrame+1)+"] = { 0, LH = { "+(outFrame+1)+", 0 }, RH = { "+(outFrame+1)+", 1 }, Flags = { StepIn = true } }\
            }\
        }";

                    savers_s.push(saver_s); 
                }
                    // *************************************************************************************************
                    loaders.push(loader); // Add loader to loaders list
                    loaders_s.push(loader_s); //
                    var exportFilePath = fileName.substring(1).slice(0, -1).replace(/\\\\/g, '\\');
                    var mediaStartTCToFrames = ((tcHH*3600)*fps)+((tcMM*60)*fps)+(tcSS*fps)+tcFF;
                    var cutInTimeCode = framesToTimeCode(mediaStartTCToFrames+(trimIn-(handles*speed)), fps);
                    var cutOutTimeCode = framesToTimeCode(mediaStartTCToFrames+(trimOut+(handles*speed)), fps);
                    if (speed != 1) {
                        exporterData += exportFilePath+"\nTime Code In: "+cutInTimeCode+"\nTime Code Out: "+cutOutTimeCode+"\nFrame In: "+(trimIn-(handles*speed))+"\nFrame Out: "+(trimOut+(handles*speed))+"\nSpeed: "+speed+"\n\n";
                    } else {
                        exporterData += exportFilePath+"\nTime Code In: "+cutInTimeCode+"\nTime Code Out: "+cutOutTimeCode+"\nFrame In: "+(trimIn-(handles*speed))+"\nFrame Out: "+(trimOut+(handles*speed))+"\n\n";
                    }
                    
                    loaderCount++; // Increase loader count
                } else {
                    // Do nothing if clip is not selected
                }
            }
        }
        // ----------------------------------------------------------------------------------------------------------------
        // Initialize Fusion Composition Begin
        var compStart = "Composition {\
    CurrentTime = 0,\
    RenderRange = { "+globalStart+", "+(globalStart+1000)+" },\
    GlobalRange = { "+globalStart+", "+(globalStart+1000)+" },\
    CurrentID = "+(loaderCount+1)+",\
    PlaybackUpdateMode = 0,\
    Version = \"Fusion 9.0.2 build 15\",\
    SavedOutputs = 2,\
    HeldTools = 0,\
    DisabledTools = 0,\
    LockedTools = 0,\
    AudioOffset = 0,\
    AutoRenderRange = true,\
    Resumable = true,\
    OutputClips = {\
    },\
    Tools = {";
        
        // Initialize Fusion Composition Ending
        var compEnd = "\n\t},\
    Prefs = {\
        Comp = {\
            Views = {\
                Right = {\
                },\
                Left = {\
                    Viewer = {\
                    },\
                },\
                View1 = {\
                },\
            },\
            FrameFormat = {\
                Rate = "+fps+",\
                GuideRatio = 1.77777777777778,\
                DepthFull = 2,\
                DepthPreview = 2,\
                DepthInteractive = 2,\
            },\
            Unsorted = {\
                GlobalEnd = 100\
            },\
            Paths = {\
            },\
            FlowView = {\
                GridSnap = true,\
            },\
            QuickTime = {\
            },\
        }\
    }\
}";
        // ----------------------------------------------------------------------------------------------------------------
        // Finalize Fusion Composition File        
        loaders   = loaders.join(",");                                           // Combine loaders        
        loaders_s = loaders_s.join(",");                                         //
        savers_s  = savers_s.join(",");                                          //
        if (timespeeds.length > 0) {                                             // If there is any time speed nodes
            timespeeds = "," + timespeeds.join(",");                               // Combine time speed nodes
        } else {                                                                 // Otherwise
            timespeeds = "";                                                     // Return empty string
        }
        if (delays_s.length > 0) {                                               // If there is any time speed nodes
            delays_s = "," + delays_s.join(",");                                   // Combine time speed nodes
        } else {                                                                 // Otherwise
            delays_s = "";                                                       // Return empty string
        }
        
        // Combine final composition string
        var comp_trimloaders   = compStart + loaders + timespeeds + compEnd;
        var comp_animatesavers = compStart + loaders_s + delays_s + timespeeds +","+ savers_s+","+ compEnd;

        // Exporting
        if (atLeastOne == 1) {                                                   // At least one loader
            if (error == 0) {                                                    // If no error

                // If export fusion comp - Trim loaders
                if (exportMethod == 0) {
                    var theFile = File.saveDialog("Where to save?",
                    "Fusion Composition:*.comp");                                // Save dialog
                    if (theFile != null) {                                       // If user didn't cancel save dialog
                        theFile.open("w","TEXT","????");                         // Open file for writing
                        theFile.encoding = "UTF-8";                              // Set encoding
                        theFile.writeln(comp_trimloaders);                       // Write comp
                        theFile.close();                                         // Close file
                        alert(loaderCount+" clips exported!");                   // Alert how many clips exported
                        if (execute == true) {                                   // If "Open Comp" checked
                            theFile.execute();                                   // Execute saved comp file
                        }
                    }

                // If export only frame data
                } else if (exportMethod == 1) {    
                    var tempFolder = Folder.temp.fsName;                         // Get temp folder path
                    var divider = getOSDivider();                                // Get divider
                    var theFile = new File(tempFolder+divider+"fusionExporterData.txt"); // Create temp text file
                    if (theFile != null) {                                       // If there is file
                        theFile.open("w","TEXT","????");
                        theFile.encoding = "UTF-8";                              // Encode
                        theFile.writeln(exporterData);
                        theFile.execute();                                       // Execute text file
                    }

                // If export fusion comp - Animate savers
                } else if (exportMethod == 2) {    
                    var theFile = File.saveDialog("Where to save?",
                    "Fusion Composition:*.comp");                                // Save dialog
                    if (theFile != null) {                                       // If user didn't cancel save dialog
                        theFile.open("w","TEXT","????");                         // Open file for writing
                        theFile.encoding = "UTF-8";                              // Set encoding
                        theFile.writeln(comp_animatesavers);                     // Write comp
                        theFile.close();                                         // Close file
                        alert(loaderCount+" clips exported!");                   // Alert how many clips exported
                        if (execute == true) {                                   // If "Open Comp" checked
                            theFile.execute();                                   // Execute saved comp file
                        }
                    }
                } 

            }
        } else {
            alert("Select at least one clip to export!");                        // Forgot something?
        }

    }
}